""" VGG16 image encoder.

Returns the convolutional features of VGG16/VGG19.
tutorial: https://keras.io/applications/.
"""
import numpy as np
import tensorflow as tf

from keras.preprocessing import image
from keras.applications.vgg16 import VGG16
from keras.applications.vgg19 import VGG19
from keras.applications.vgg19 import preprocess_input
from keras.models import Model


class VGG16ImageEncoder:
    """ VGG16 Image Encoder with default 'fc1' top layer. """
    def __init__(self, top_layer='fc1', print_summary=False):
        base_model = VGG16(weights='imagenet')
        self.model = Model(inputs=base_model.input, outputs=base_model.get_layer(top_layer).output)
        self.graph = tf.get_default_graph()
        if print_summary:
            self.model.summary()

    def batch_encode(self, img_paths):
        """ Batch encode images into VGG16 embedding features. """
        features = []
        for img_path in img_paths:
            features.append(self.encode(img_path))
        return features

    def encode(self, img_path):
        """ Encode an image into VGG16 embedding features. """
        # Running predict on webservice, requires fetching tf.graph
        # See: https://github.com/keras-team/keras/issues/2397
        with self.graph.as_default():
            img = image.load_img(img_path, target_size=(224, 224))
            img_data = image.img_to_array(img)
            img_data = np.expand_dims(img_data, axis=0)
            img_data = preprocess_input(img_data)
            feature = self.model.predict(img_data)
            feature_np = np.array(feature)
            return feature_np.flatten()


class VGG19ImageEncoder:
    """ VGG19 Image Encoder with default 'fc1' top layer. """
    def __init__(self, top_layer='fc1', print_summary=False):
        base_model = VGG19(weights='imagenet')
        self.model = Model(inputs=base_model.input, outputs=base_model.get_layer(top_layer).output)
        self.graph = tf.get_default_graph()
        if print_summary:
            self.model.summary()

    def batch_encode(self, img_paths):
        """ Batch encode images into VGG19 embedding features. """
        features = []
        for img_path in img_paths:
            features.append(self.encode(img_path))
        return features

    def encode(self, img_path, flattened=False):
        """ Encode an image into VGG19 embedding features. """
        # Running predict on webservice, requires fetching tf.graph
        # See: https://github.com/keras-team/keras/issues/2397
        with self.graph.as_default():
            img = image.load_img(img_path, target_size=(224, 224))
            img_data = image.img_to_array(img)
            img_data = np.expand_dims(img_data, axis=0)
            img_data = preprocess_input(img_data)
            feature = self.model.predict(img_data)
            feature_np = np.array(feature)
            return feature_np.flatten() if flattened else feature_np



""" Hawking CLI.

Hawking Search Engine Command Line Interface (CLI).
Copyright (C) 2019 Peach Inc. All rights reserved.
"""
import os
import click
import json
import shutil
import subprocess
import time

import grpc

from hawking.common.color import *


_HAWKING_BASE_IMAGE = 'hawking-base:0.1'
_BASE_PIP_PACKAGES = 'hawking==0.1.55 hawking_proto==0.1.15 faiss-cpu'


def bold(skk):
    print_color(skk, bold=True, color=LIGHT_YELLOW)


def info(skk):
    print_color(skk, color=LIGHT_CYAN)


def hilite(skk):
    print_color(skk, underline=True, color=LIGHT_GREEN)


def shell_execute(cmds):
    for cmd in cmds:
        info("executing: `{}`".format(cmd))
        os.system(cmd)


def get_container_ip(tag):
    cmd = "docker ps | awk '{ if ($2 == \"%s\") { print $1 }}' | xargs docker container inspect" % (tag)
    result = subprocess.check_output(cmd, shell=True)
    r = json.loads(result)
    ip_addresses = []
    for i in r:
        ip = i['NetworkSettings']['IPAddress']
        ip_addresses.append(ip)
        print(ip)
    return ip_addresses


def kill_container(tag):
    cmd = "docker ps | awk '{ if ($2 == \"%s\") { print $1 } }' | xargs docker kill" % (tag)
    result = subprocess.check_output(cmd, shell=True)
    print(result.decode('utf-8'))


@click.group()
def cli():
    pass


def build_base_image(dockerfile='hawking.dockerfile.tmp'):
    def create_docker_file(p_dockerfile):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM hawking-rte:0.1

# Separating this as an optimization, since pytorch is a large .whl (768mb)
RUN apt-get update \
    && . /root/miniconda3/etc/profile.d/conda.sh \
    && conda activate hawking \
    && pip install torch pytorch-transformers

RUN . /root/miniconda3/etc/profile.d/conda.sh \
    && conda activate hawking \
    && pip install {0}
""".format(_BASE_PIP_PACKAGES))

    create_docker_file(dockerfile)
    cmds = [
        'docker build -f {} -t {} .'.format(dockerfile, _HAWKING_BASE_IMAGE),
        'cat {0}'.format(dockerfile),
        'rm -f {0}'.format(dockerfile),
    ]
    shell_execute(cmds)


def deploy_bot_as_local_docker_daemon(token, dockerfile, tag, build_only=False):
    def create_docker_file(p_dockerfile):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM {0}

CMD . /root/miniconda3/etc/profile.d/conda.sh \
  && conda activate hawking \
  && hawking start bot > /tmp/bot.log 2> /tmp/bot.err
""".format(_HAWKING_BASE_IMAGE))
    create_docker_file(dockerfile)

    cmds = [
        'docker build -f {} -t {} .'.format(dockerfile, tag),
        'docker run -d --rm -v /tmp:/tmp -e BOT_TOKEN="{0}" {1}'.format(token, tag),
        'cat {0}'.format(dockerfile),
        'rm -f {0}'.format(dockerfile),
    ]

    shell_execute([cmds[0], cmds[2], cmds[3]] if build_only else cmds)


def deploy_secretary_as_local_docker_daemon(token, dockerfile, tag, build_only=False):
    def create_docker_file(p_dockerfile):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM {0}

CMD . /root/miniconda3/etc/profile.d/conda.sh \
  && conda activate hawking \
  && hawking start secretary > /tmp/secretary.log 2> /tmp/secretary.err
""".format(_HAWKING_BASE_IMAGE))
    create_docker_file(dockerfile)

    cmds = [
        'docker build -f {} -t {} .'.format(dockerfile, tag),
        'docker run -d --rm -v /tmp:/tmp -e BOT_TOKEN="{0}" {1}'.format(token, tag),
        'cat {0}'.format(dockerfile),
        'rm -f {0}'.format(dockerfile),
    ]

    shell_execute([cmds[0], cmds[2], cmds[3]] if build_only else cmds)


def deploy_caption_serving_as_local_docker_daemon(data_root, args, dockerfile, tag, port, build_only=False):
    def create_docker_file(p_dockerfile):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM {0}

CMD . /root/miniconda3/etc/profile.d/conda.sh \
  && conda activate hawking \
  && hawking start caption $ARGS > /tmp/caption.log 2> /tmp/caption.err
""".format(_HAWKING_BASE_IMAGE))
    create_docker_file(dockerfile)
    cmds = [
        'docker build -f {} -t {} .'.format(dockerfile, tag),
        'docker run -d --rm -p {0}:{0} -v {1}:/data/ -v /tmp:/tmp -v $PWD:/host -e ARGS="{2}" {3}'.format(
            port, data_root, args, tag),
        'cat {0}'.format(dockerfile),
        'rm -f {0}'.format(dockerfile),
    ]
    shell_execute([cmds[0], cmds[2], cmds[3]] if build_only else cmds)


def deploy_fx_serving_as_local_docker_daemon(data_dir, args, dockerfile, tag, port, build_only=False):
    def create_docker_file(p_dockerfile):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM {0}

CMD . /root/miniconda3/etc/profile.d/conda.sh \
  && conda activate hawking \
  && hawking start fx $ARGS > /tmp/fx.log 2> /tmp/fx.err
""".format(_HAWKING_BASE_IMAGE))
    create_docker_file(dockerfile)

    cmds = [
        'docker build -f {} -t {} .'.format(dockerfile, tag),
        'docker run -d --rm -p {0}:{0} -v {1}:/data/ -v /tmp:/tmp -v $PWD:/host -e ARGS="{2}" {3}'.format(
            port, data_dir, args, tag),
        'cat {0}'.format(dockerfile),
        'rm -f {0}'.format(dockerfile),
    ]

    shell_execute([cmds[0], cmds[2], cmds[3]] if build_only else cmds)


def deploy_sx_serving_as_local_docker_daemon(index_dir, index_name, args, dockerfile, port, tag, build_only=False):
    def create_docker_file(p_dockerfile):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM {0}

RUN mkdir -p /app/sptag

COPY $HAWKING_ROOT/src/cli/hawking/serve/sptag_server.py /app/sptag

WORKDIR /app/sptag

CMD . /root/miniconda3/etc/profile.d/conda.sh \
  && conda activate hawking \
  && /bin/bash -c "python /app/sptag/sptag_server.py --index-path /data/$INDEX_NAME $ARGS"
""".format(_HAWKING_BASE_IMAGE))

    create_docker_file(dockerfile)

    cmds = [
        'docker build -f {} -t {} .'.format(dockerfile, tag),
        'docker run -d --rm -p {0}:{0} -v {1}:/data/ -e INDEX_NAME="{2}" -e ARGS="{3}" {4}'.format(
            port, index_dir, index_name, args, tag),
        'cat {0}'.format(dockerfile),
        'rm -f {0}'.format(dockerfile),
    ]

    shell_execute([cmds[0], cmds[2], cmds[3]] if build_only else cmds)


def deploy_bert_serving_as_local_docker_daemon(model_dir, model_name, args, dockerfile, tag, build_only=False):
    def create_docker_file(p_dockerfile):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM {0}

CMD . /root/miniconda3/etc/profile.d/conda.sh \
  && conda activate hawking \
  && echo BERT_ARGS=$BERT_ARGS \
  && /bin/bash -c "$(which bert-serving-start) -model_dir=/data/{1} $BERT_ARGS"
""".format(_HAWKING_BASE_IMAGE, model_name))

    create_docker_file(dockerfile)

    cmds = [
        'docker build -f {} -t {} .'.format(dockerfile, tag),
        'docker run -d -p 5555:5555 -p 5556:5556 -v {0}:/data/ -e BERT_ARGS="{1}" {2}'.format(model_dir, args, tag),
        'cat {0}'.format(dockerfile),
        'rm -f {0}'.format(dockerfile),
    ]

    shell_execute([cmds[0], cmds[2], cmds[3]] if build_only else cmds)


def deploy_sx_indexer_as_local_docker_daemon(index_dir, index_name, service_ini, dockerfile, port, tag, build_only=False):
    """ Deploy SX backend server local docker. """
    def create_service_ini(p_output, p_index_name, p_port, p_max_result_count=5):
        """ Do a set of regexes to replace default values with overrides. """
        service_ini_tmpl = "$HAWKING_ROOT/src/cli/hawking/sptag/service.ini"
        _cmds = [
            ("sed -e 's/ListenPort=.*/ListenPort={0}/g' {1}".format(p_port, service_ini_tmpl)),
            ("sed -e 's/DefaultMaxResultNumber=.*/DefaultMaxResultNumber={0}/g'".format(p_max_result_count)),
            ("sed -e 's/merge_2006-10_2007-04-08_2018-05/{0}/g'".format(p_index_name))
        ]

        os.system("cp {0} {1}".format(service_ini_tmpl, p_output))
        _cmd_string = _cmds[0]
        for _cmd in _cmds[1:]:
            _cmd_string += " | " + _cmd
        _cmd_string += " > " + p_output
        print(_cmd_string)
        os.system(_cmd_string)

    def create_docker_file(p_dockerfile, p_service_ini, p_index_dir, p_index_name):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM hawking-rte:0.1
COPY $HAWKING_ROOT/src/cli/hawking/sptag/server /app/sptag/
COPY {2} /app/sptag/service.ini
ENTRYPOINT ["/bin/bash", "-c", "/app/sptag/server -m socket -c /app/sptag/service.ini"]
""".format(p_index_dir, p_index_name, p_service_ini))

    hawking_root = os.environ.get('HAWKING_ROOT')
    if hawking_root is None:
        assert False, "$HAWKING_ROOT is not set. Aborting ..."
    assert os.path.exists("{0}/src/cli/hawking/sptag".format(hawking_root)), \
        "`{0}/src/cli/hawking/sptag` does not exist. Aborting ...".format(hawking_root)
    assert os.path.exists("{0}/{1}".format(index_dir, index_name)), \
        "Index `{0}/{1}` does not exist. Aborting ...".format(index_dir, index_name)
    create_service_ini(service_ini, index_name, port)
    create_docker_file(dockerfile, service_ini, index_dir, index_name)

    cmds = [
        'docker build -f {0} -t {1} .'.format(dockerfile, tag),
        'docker run -d -p {0}:{0} -v {1}:/data {2}'.format(port, index_dir, tag),
        'cat {}'.format(dockerfile),
        'rm -f {0} && rm -f {1}'.format(dockerfile, service_ini)
    ]

    shell_execute([cmds[0], cmds[2], cmds[3]] if build_only else cmds)


def deploy_queue_as_local_docker_daemon(queue_dir, args, dockerfile, port, tag, build_only=False):
    # TODO(oleg): refactor similar to fx and caption
    def create_docker_file(p_dockerfile):
        with open(p_dockerfile, 'w') as f:
            f.write("""
FROM {0}

RUN mkdir -p /app/queue

COPY $HAWKING_ROOT/src/cli/hawking/serve/queue_server.py /app/queue

WORKDIR /app/queue

CMD . /root/miniconda3/etc/profile.d/conda.sh \
  && conda activate hawking \
  && /bin/bash -c "python /app/queue/queue_server.py --queue-path /data $ARGS"
""".format(_HAWKING_BASE_IMAGE))
    create_docker_file(dockerfile)

    cmds = [
        'docker build -f {} -t {} .'.format(dockerfile, tag),
        'docker run -d -p {0}:{0} -v {1}:/data/ -e ARGS="{2}" {3}'.format(
            port, queue_dir, args, tag),
        'cat {0}'.format(dockerfile),
        'rm -f {0}'.format(dockerfile),
    ]

    shell_execute([cmds[0], cmds[2], cmds[3]] if build_only else cmds)


@cli.group()
def deploy():
    """ Deploy command group. """
    build_base_image()
    pass


@deploy.command()
@click.argument('data_root')
@click.argument('index_name')
@click.option('--grpc-port', help='Server port.', default=50001)
@click.option('--tag', '-t', help='Frontend container tag.', default='caption-serving:0.1')
@click.option('--worker-threads', '-w', help='Max worker threads.', default=10)
@click.option('--build-only/--no-build-only', help='Build docker image only.', default=False)
def caption(data_root, index_name, grpc_port, tag, worker_threads, build_only):
    """ Deploy Caption serving node. """
    print('[Deploy Caption index serving node]')
    assert os.path.exists(data_root), "data_root: {} does not exist.".format(data_root)
    index_path = os.path.join(data_root, index_name)
    if not index_path.endswith('.caption'):
        index_path = '{}.caption'.format(index_path)
    assert os.path.exists(index_path), "--index-path `{}` does not exist. Aborting.".format(index_path)
    args = '--index-path /data/{} --worker-threads {} --grpc-port {}'.format(index_name, worker_threads, grpc_port)
    deploy_caption_serving_as_local_docker_daemon(data_root, args, 'caption-serving.dockerfile.tmp', tag, grpc_port,
                                                  build_only=build_only)


@deploy.command()
@click.option('--token', help='BOT_TOKEN value.', default='701884073:AAF9rkmy73sq6U3QPbZxX5JzyfkDqhP-5xQ')
@click.option('--tag', '-t', help='Container tag.', default='bot:0.1')
@click.option('--build-only/--no-build-only', help='Build docker image only.', default=False)
def bot(token, tag, build_only):
    """ Deploy telegram bot. """
    print('[Deploy Telegram Bot]')
    deploy_bot_as_local_docker_daemon(token, 'telegram_bot.dockerfile.tmp', tag, build_only=build_only)


@deploy.command()
@click.option('--token', help='BOT_TOKEN value.', default='701884073:AAF9rkmy73sq6U3QPbZxX5JzyfkDqhP-5xQ')
@click.option('--tag', '-t', help='Container tag.', default='bot:0.1')
@click.option('--build-only/--no-build-only', help='Build docker image only.', default=False)
def secretary(token, tag, build_only):
    """ Deploy secretary bot. """
    print('[Deploy Secretary Bot]')
    deploy_secretary_as_local_docker_daemon(token, 'secretary_bot.dockerfile.tmp', tag, build_only=build_only)


@deploy.command()
@click.argument('data_root')
@click.argument('index_name')
@click.option('--grpc-port', help='Server port.', default=50000)
@click.option('--caption-host', help='Caption server host url/ip.', default='127.0.0.1')
@click.option('--caption-port', help='Caption server port.', default=50001)
@click.option('--indexer-host', help='Indexer server host url/ip.', default='127.0.0.1')
@click.option('--indexer-port', help='Indexer server port.', default=5000)
@click.option('--encoder-config', help='Encoder config file.', default='/tmp/encoder.json')
@click.option('--tag', '-t', help='Container tag.', default='fx-serving:0.1')
@click.option('--worker-threads', '-w', help='Max worker threads.', default=10)
@click.option('--build-only/--no-build-only', help='Build docker image only.', default=False)
def fx(data_root, index_name, grpc_port, caption_host, caption_port, indexer_host, indexer_port, encoder_config, tag,
       worker_threads, build_only):
    """ Deploy FX index serving node. """
    print('[Deploy FX Index Serving Nodes]')
    assert os.path.exists(data_root), "data_root: {} does not exist.".format(data_root)
    index_path = os.path.join(data_root, index_name)
    if not index_path.endswith('.fvec'):
        index_path = '{}.fvec'.format(index_path)
    assert os.path.exists(index_path), "--index-path `{}` does not exist. Aborting.".format(index_path)
    assert os.path.exists(encoder_config), "--encoder_config: {} does not exist.".format(encoder_config)
    print('Copying {} to {}'.format(encoder_config, data_root))
    try:
        shutil.copy(encoder_config, data_root)
    except Exception as e:
        print(e)
        pass
    args = ('--index-path=/data/{} --encoder-config=/data/{} --indexer-host={} --indexer-port={} '
            '--caption-host={} --caption-port={} --worker-threads={} --grpc-port {}'.format(
        index_name, os.path.basename(encoder_config), indexer_host, indexer_port, caption_host, caption_port,
        worker_threads, grpc_port))
    deploy_fx_serving_as_local_docker_daemon(data_root, args, 'fx-serving.dockerfile.tmp', tag, grpc_port,
                                             build_only=build_only)


@deploy.command()
@click.argument('model_dir')
@click.argument('model_name')
@click.option('--num-worker', '-w', help='Number of bert serving worker.', default=1)
@click.option('--max-batch-size', '-b', help='Bert serving max batch size.', default=16)
@click.option('--tag', '-t', help='Container tag', default='bert-serving:0.1')
@click.option('--build-only/--no-build-only', help='Build docker image only.', default=False)
def bert(model_dir, model_name, num_worker, max_batch_size, tag, build_only):
    """ Deploy BERT serving node. """
    print('[Deploy BERT serving node]')
    deploy_bert_serving_as_local_docker_daemon(
        model_dir, model_name, '-num_worker={} -max_batch_size={}'.format(num_worker, max_batch_size),
        'bert-serving.dockerfile.tmp', tag, build_only=build_only)


@deploy.command()
@click.argument('index_dir')
@click.argument('index_name')
@click.option('--encoder-host', help='Encoder host url/ip.', default='127.0.0.1')
@click.option('--encoder-port', help='Encoder host port.', default=5555)
@click.option('--grpc-port', help='Server port.', default=50000)
@click.option('--indexer-host', help='Indexer server host url/ip.', default='127.0.0.1')
@click.option('--indexer-port', help='Indexer server port.', default=5000)
@click.option('--backend-tag', '-bt', help='Indexer container tag.', default='sx-indexer:0.1')
@click.option('--frontend-tag', '-ft', help='Frontend container tag.', default='sx-serving:0.1')
@click.option('--worker-threads', '-w', help='Max worker threads.', default=10)
@click.option('--build-only/--no-build-only', help='Build docker image only.', default=False)
def sx(index_dir, index_name, encoder_host, encoder_port, grpc_port, indexer_host, indexer_port, backend_tag,
       frontend_tag, worker_threads, build_only):
    """ Deploy SX index serving node. """
    print('[Deploy SX index serving node]')
    deploy_sx_indexer_as_local_docker_daemon(
        index_dir, index_name,
        'service.ini.tmp', 'sx-indexer.dockerfile.tmp', indexer_port, backend_tag, build_only=build_only)

    indexer_ip_addresses = []
    counter = 0
    while counter < 10:
        indexer_ip_addresses = get_container_ip('sx-indexer:0.1')
        if len(indexer_ip_addresses) == 0:
            print('Waiting for indexer to start ...')
            time.sleep(3)
        else:
            break
        counter += 1

    if counter == 10:
        print('Failed to start SX indexer')
        sys.exit(1)

    args = '--encoder-host {} --encoder-port {} --indexer-host {} --indexer-port {} --worker-threads {} --grpc-port {}'.format(
        encoder_host, encoder_port, indexer_ip_addresses[0], indexer_port, worker_threads, grpc_port)

    deploy_sx_serving_as_local_docker_daemon(
        index_dir, index_name, args, 'sx-serving.dockerfile.tmp', grpc_port, frontend_tag, build_only=build_only)


@deploy.command()
@click.argument('queue_dir')
@click.option('--grpc-port', help='Server port.', default=45000)
@click.option('--tag', '-t', help='Container tag.', default='simple-queue:0.1')
@click.option('--worker-threads', '-w', help='Max worker threads.', default=10)
@click.option('--build-only/--no-build-only', help='Build docker image only.', default=False)
def queue(queue_dir, grpc_port, tag, worker_threads, build_only):
    """ Deploy Queue node. """
    print('[Deploy Queue Nodes]')
    args = ('--worker-threads={} --grpc-port {}'.format(
            worker_threads, grpc_port))
    deploy_queue_as_local_docker_daemon(
        queue_dir, args, 'queue.dockerfile.tmp', grpc_port, tag, build_only=build_only)


@cli.group()
def start():
    pass


@start.command()
def bot():
    """ Deploy Hawking Telegram Bot. """
    # TODO(oleg): deploy this as docker container
    from hawking.bot.telegram import main
    main()


@start.command()
def secretary_bot():
    """ Deploy Secretary Bot. """
    # TODO(oleg): deploy this as docker container
    from hawking.bot.secretary import main
    main()


@start.command()
@click.option('--port', '-p', help="Listening port", default=8080)
def frontend(port):
    """ Deploy Hawking front-end. """
    # TODO(oleg): deploy this as docker container
    from hawking.frontend.application import main
    main(port)


@start.command()
@click.option('--index-path', help='Path to index', default=None)
@click.option('--grpc-port', help='Server port.', default=50000)
@click.option('--caption-host', help='Caption server host url/ip.', default='127.0.0.1')
@click.option('--caption-port', help='Caption server port.', default=50001)
@click.option('--indexer-host', help='Indexer server host url/ip.', default='127.0.0.1')
@click.option('--indexer-port', help='Indexer server port.', default=5000)
@click.option('--encoder-config', help='Encoder config file.', default='/tmp/encoder.json')
@click.option('--worker-threads', '-w', help='Max worker threads.', default=10)
@click.option('--run-bot/--no-run-bot', help='Run telegram bot.', default=False)
def fx(index_path, grpc_port, caption_host, caption_port, indexer_host, indexer_port, encoder_config, worker_threads,
       run_bot):
    """ Start fx serving. """
    from hawking.serve.faiss_server import serve as fx_serve
    with open(encoder_config, 'r') as f:
        encoder_config = json.load(f)
    if not index_path.endswith('.fvec'):
        index_path = '{}.fvec'.format(index_path)
    assert os.path.exists(index_path), "--index-path `{}` does not exist. Aborting.".format(index_path)
    fx_serve(index_path, indexer_host, indexer_port, caption_host, caption_port, worker_threads, grpc_port,
             encoder_config, run_bot)


@start.command()
@click.option('--index-path', help='Path to index', default=None)
@click.option('--grpc-port', help='Server port.', default=50001)
@click.option('--worker-threads', '-w', help='Max worker threads.', default=10)
def caption(index_path, grpc_port, worker_threads):
    """ Start caption serving. """
    from hawking.serve.caption_server import serve as caption_serve
    if not index_path.endswith('.caption'):
        index_path = '{}.caption'.format(index_path)
    assert os.path.exists(index_path), "--index-path `{}` does not exist. Aborting.".format(index_path)
    caption_serve(index_path, worker_threads, grpc_port)


@cli.command()
@click.argument('doc_ids', nargs=-1)
@click.option('--server-ip', '-s', help="Remote caption host url/ip", default='127.0.0.1')
@click.option('--port', '-p', help="Remote caption server port", default=50001)
def caption(doc_ids, server_ip, port):
    """ Query Hawking caption server. """
    from hawking_proto import hawking_pb2, hawking_pb2_grpc

    with grpc.insecure_channel('{}:{}'.format(server_ip, port)) as channel:
        stub = hawking_pb2_grpc.CaptionStub(channel)
        request = hawking_pb2.SnippetRequest(max_length=200, prettify=True)
        for doc_id in doc_ids:
            request.doc_ids.append(int(doc_id))

        bold('DocIds: "{}"'.format(','.join(doc_ids)))
        st = time.time()
        response = stub.Get(request)
        bold('Caption results:\n')
        for snippet in response.snippets:
            hilite('"%s"' % snippet.title)

        et = (time.time() - st) * 1000.0
        bold('\nelapsed time: %d ms' % et)


@cli.command()
@click.option('--entity-list', '-list', help='File containing list of entities', default=None)
@click.option('--batch-size', '-b', help='Size of batch encoding request', default=500)
@click.option('--doc-count', '-n', help='Number of docs', default=100)
@click.option('--output', '-o', help='Output filename', default='hawking.em')
@click.option('--dimension', '-dim', help='Embedding dimension', default=768)
@click.option('--random-index/--no-random-index', '-r/-no-r', help='Generate random index', default=False)
@click.option('--encoder-config', help="Encoder config", default='/tmp/encoder.json')
def batch_index(entity_list, batch_size, doc_count, output, dimension, random_index, encoder_config):
    """ Create batch index (offline).

    When --random is specified, generate random index with random embedding.
    """
    from hawking.encoder import create_encoder

    if not random_index:
        assert entity_list is not None, "When --random-index is False, --entity-list argument is mandatory."
    dimension = int(dimension)
    from random import random, shuffle

    class DocumentIterator:
        def __init__(self, source_file):
            source_file = source_file
            handle = None

        def next(self):
            raise Exception('You are calling an abstract class method. Please use the derived class next().')

        source_file = ''
        _handle = None

    class TextIterator(DocumentIterator):
        def __init__(self, source_file):
            self.source_file = source_file
            self._handle = open(self.source_file, 'r')

        def next(self):
            for line in self._handle:
                yield line.strip()
            self._handle.close()

    def gen_rand_metadata():
        s = 'abc 123 ghi j#@ m$o,.?'
        l = list(s)
        shuffle(l)
        return ''.join(l)

    def gen_rand_index(p_output, p_metadata_fn, p_dim):
        """ Generate random sptag index

        format:
        <text-0>\t<v0>,...,<v-dim-1>\n
        ...
        <text-n>\t<v0>,...,<v-dim-1>\n
        """
        with open(p_output, 'w') as f:
            for i in range(doc_count):
                f.write(p_metadata_fn())
                f.write('\t')
                f.write('%f' % random())
                for n in range(p_dim - 1):
                    f.write('|')
                    f.write('%f' % random())
                f.write('\n')
        return doc_count

    def write_record(p_out, p_phrases, p_vectors, p_dim):
        """ Write record (<phrase>\t<vector>\n) to a file. """
        for i, v in enumerate(p_vectors):
            p_out.write(p_phrases[i])
            p_out.write('\t')
            assert v.size > 0 and v.size == p_dim, 'Embedding dimension does not match. Expected:{}. Actual:{}'.format(p_dim, v.size)
            value = '|'.join([repr(a) for a in v])
            p_out.write(value)
            p_out.write('\n')

    def gen_embedding_index(p_output, p_doc_iter, p_dim, p_encoder, p_batch_size=100):
        """ Generate sptag index from document list

        format:
        <text-0>\t<v0>,...,<v-p_dim-1>\n
        ...
        <text-n>\t<v0>,...,<v-p_dim-1>\n
        """
        phrases = []
        count = 0
        with open(output, 'w') as f:
            start = time.time()
            for phrase in p_doc_iter.next():
                if phrase == '' or len(phrase) == 0:
                    continue
                phrases.append(phrase)
                if len(phrases) < p_batch_size:
                    continue
                count += len(phrases)
                vectors = p_encoder.batch_encode(phrases)
                assert len(vectors) == len(phrases)
                write_record(f, phrases, vectors, p_dim)
                elapsed_secs = time.time() - start
                info('Progress: %d docs. Elapsed: %.3f s. Rate: %.2f dps ...' % (count, elapsed_secs, len(phrases)/elapsed_secs))
                vectors = []
                phrases = []
                start = time.time()

            if len(phrases) > 0:
                count += len(phrases)
                start = time.time()
                info('Indexing remaining {} docs'.format(len(phrases)))
                vectors = p_encoder.batch_encode(phrases)
                assert len(vectors) == len(phrases)
                write_record(f, phrases, vectors, p_dim)
                elapsed_secs = time.time() - start
                info('Progress: %d docs. Elapsed: %.3f s. Rate: %.2f dps ...' % (count, elapsed_secs, len(phrases)/elapsed_secs))
                vectors = []
                phrases = []
        return count

    start = time.time()
    if random_index:
        info('Generate random embedding ...')
        total_entity_count = gen_rand_index(output, gen_rand_metadata, dimension)
    else:
        info('Generate embedding file from:`{}`...'.format(entity_list))
        entity_iter = TextIterator(entity_list)
        print('Encoder Config:{}'.format(encoder_config))
        with open(encoder_config, 'r') as f:
            encoder_config = json.load(f)
        encoder = create_encoder(config=encoder_config)
        if encoder_config['entity_type'] == 'text':
            total_entity_count = gen_embedding_index(output, entity_iter, dimension, encoder, batch_size)
        elif encoder_config['entity_type'] == 'image':
            total_entity_count = gen_embedding_index(output, entity_iter, dimension, encoder, batch_size)
        else:
            raise ValueError('-t {} is not supported!'.format(encoder_config['entity_type']))
    print('Completed building index.')
    print(' Output       : %s' % output)
    print(' Total #docs  : %d' % total_entity_count)
    print(' Elapsed time : %.3f secs' % (time.time() - start))
    print(' Batch size   : %d' % batch_size)


@cli.group()
def text():
    """ Text command group. """
    pass


@text.command()
@click.argument('index')
@click.option('--embedding', 'e', help="A single atom embedding, format: <atom>\t<v0> <v1> ... <vK>", default=None)
@click.option('--embedding_file', '-ef', help='Embedding file, rows of atoms embeddings', default=None)
def index(index, embedding, embedding_file):
    """ Add atom embedding(s) to Hawking index in online manner. """
    print('[Updating embedding index]')
    if (embedding is None
            and embedding_file is None):
        raise Exception('Either --embedding or --embedding_file must be specified.')


@text.command()
@click.argument('phrases', nargs=-1)
@click.option('--encoder-host', '-e', help="Remote encoder host url/ip", default='127.0.0.1')
@click.option('--encoder-port', '-p', help="Remote encoder port", default=5555)
@click.option('--encoder-type', help="Encoder type.", default="universal-sentence-encoder")
@click.option('--weights', help="Pre-trained weights uri, it can be name or uri.",
              default="https://tfhub.dev/google/universal-sentence-encoder-multilingual-qa/1")
def encode(phrases, encoder_host, encoder_port, encoder_type, weights):
    """ Encode phrase(s) into vector embedding.

    encoder-type: online-bert
    weights: [ignored]

    encoder-type: bert
    weights: https://huggingface.co/pytorch-transformers/pretrained_models.html

    encoder-type: universal-sentence-encoder
    weights:
    - https://tfhub.dev/google/universal-sentence-encoder/2
    - https://tfhub.dev/google/universal-sentence-encoder-large/3
    - https://tfhub.dev/google/universal-sentence-encoder-lite/2
    - https://tfhub.dev/google/universal-sentence-encoder-multilingual/1
    - https://tfhub.dev/google/universal-sentence-encoder-multilingual-qa/1
    """
    from hawking.encoder import create_encoder
    encoder = create_encoder(config={
        "entity_type": "text",
        "encoder_type": encoder_type,
        "weights": weights,
        "encoder_host": encoder_host,
        "encoder_port": int(encoder_port),
    })
    for p in phrases:
        v = encoder.encode(p)
        print(v)


@text.command()
@click.argument('phrases', nargs=-1)
@click.option('--server-ip', '-s', help="Remote Hawking host url/ip", default='127.0.0.1')
@click.option('--port', '-p', help="Remote Hawking server port", default=50000)
@click.option('--result-count', '-n', help="Result count", default=10)
def query(phrases, server_ip, port, result_count):
    """ Query Hawking index server. """
    from hawking_proto import hawking_pb2, hawking_pb2_grpc

    with grpc.insecure_channel('{}:{}'.format(server_ip, port)) as channel:
        stub = hawking_pb2_grpc.SearchStub(channel)
        for phrase in phrases:
            bold('Query: "{}"'.format(phrase))
            st = time.time()
            request = hawking_pb2.SearchRequest(max_results=result_count, query=phrase,
                                                type=hawking_pb2.SearchRequest.EntityType.TEXT)
            response = stub.Get(request)
            bold('Search results:\n')
            rank = 1
            for result in response.results:
                hilite('%d. "%s" [docid:%s, score:%.6f]' % (rank, result.document.body, result.document.uuid, result.score))
                rank += 1

            et = (time.time() - st) * 1000.0
            bold('\nelapsed time: %d ms' % et)


@cli.group()
def image():
    """ Image command group. """
    pass


@image.command()
@click.argument('uris', nargs=-1)
@click.option('--arch', '-a', help="Neural Net Architecture: ['resnet50', 'vgg16', 'vgg19', 'inceptionresnetv2']", default='resnet50')
@click.option('--top-layer', '-t', help="Top layer output. resnet50:['avg_pool', 'fc1000'], "
              "inceptionresnetv2:['predictions','avg_pool']", default='fc1000')
@click.option('--print-summary/--no-print-summary', '-p/-no-p', help="To print model summary or not", default=True)
def encode(uris, arch, top_layer, print_summary):
    """ Encode image into embedding features. """
    from hawking.encoder import create_encoder
    enc = create_encoder('image', arch, config={"top_layer": top_layer, "print_summary": print_summary})

    for uri in uris:
        v = enc.encode(uri)
        print(v)
        print(v.shape)


@image.command()
@click.argument('uris', nargs=-1)
@click.option('--port', '-p', help="Port of the hawking server.", default=50000)
@click.option('--result-count', '-n', help="Result count.", default=10)
@click.option('--server-ip', '-s', help="Server IP address.", default='127.0.0.1')
def find(uris, port, result_count, server_ip):
    """ Query Hawking index server. """
    from hawking_proto import hawking_pb2, hawking_pb2_grpc

    with grpc.insecure_channel('{}:{}'.format(server_ip, port)) as channel:
        stub = hawking_pb2_grpc.SearchStub(channel)
        for uri in uris:
            bold('Find: "{}"'.format(uri))
            st = time.time()
            request = hawking_pb2.SearchRequest(max_results=result_count, query=uri,
                                                type=hawking_pb2.SearchRequest.EntityType.IMAGE)
            response = stub.Get(request)
            bold('Search results:\n')
            rank = 1
            for result in response.results:
                hilite('%d. "%s" [docid:%s, score:%.6f]' % (rank, result.document.body, result.document.uuid, result.score))
                rank += 1

            et = (time.time() - st) * 1000.0
            bold('\nelapsed time: %d ms' % et)


@cli.command()
def port_table():
    """ List port table. """
    print('IndexServing:5000')
    print('BertServing:5555')
    print('FX-Serving:50000')
    print('SX-Serving:50000')
    print('CaptionServing:50001')


@cli.command()
@click.argument('tags', nargs=-1)
def get_ip(tags):
    """ Get IP Address of containers. """
    for tag in tags:
        get_container_ip(tag)


@cli.command()
@click.argument('tags', nargs=-1)
def kill(tags):
    """ Kill containers by tag """
    for tag in tags:
        kill_container(tag)


def main():
    cli()


if __name__ == '__main__':
    main()



""" Hawking Telegram Bot.

"""
import logging
import os
import random
import sys

from hawking.frontend.engine.default import engine
from telegram import Sticker
from telegram.ext import Updater, CommandHandler, MessageHandler, Filters

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
log.addHandler(handler)


def start(update, context):
    log.info("Conversation started!")
    update.message.reply_text('Welcome, to next gen search engine')


def help(update, context):
    log.info("Help requested!")
    update.message.reply_text("I'm Hawking, `/ask` <anything>")


def handle_query(update, context):
    if update.message == None:
        return

    if update.message.text == None:
        return

    namespace = "chat_" + str(update.message.chat.id)
    query = update.message.text
    if query.startswith("/ask"):
        query = query[len("/ask"):]
        query = query.strip()

    log.info('search within "%s" namespace for "%s"', namespace, query)

    if query.startswith("http"):
        entity_type = 'image'
    else:
        entity_type = 'text'
    results = engine.search(namespace, 3, query, entity_type)

    chat_id = update.message.chat.id
    if entity_type == 'image':
        # For testing
        # results = [
        #     ("https://storage.googleapis.com/imagestash001/906802_3a8ac174-a9cc-4e46-88b3-bfda75941d9f_1076_1061.png", 0),
        #     ("https://storage.googleapis.com/imagestash001/23924271_8b9fb814-228b-4d4e-b51c-3e7943cd1c43_1080_1080.jpg", 1),
        #     ("https://storage.googleapis.com/imagestash001/122138742_71ddecc0-6531-4007-a622-91b076c0b602_1208_1208.jpg", 2)
        # ]
        for result in results:
            context.bot.send_photo(chat_id, result[0])
    else:
        sticker_ids = [
            'CAADAgADYwADJxRJC5prZI7aN9pbFgQ',
            'CAADAgADaAADJxRJC5_7BxWEYU1nFgQ',
            'CAADAwADVQADCZURAAFEBwkgoXPB0BYE',
            'CAADAgADKQcAAmMr4gnhM9ccDb2hKBYE',
            'CAADAgADmAcAAmMr4gkxuVuNCmpEdhYE',
            'CAADAgAD0wgAAgi3GQLjMLXpwKu9HxYE',
            'CAADAgADaAMAAn7yxQzzQ-DbolYIjxYE',
            'CAADAwADVQADCZURAAFEBwkgoXPB0BYE',
        ]
        context.bot.send_sticker(chat_id, sticker=random.choice(sticker_ids))
        message = '*Query*: *"{}"*\n\n'.format(query)
        rank = 1
        for result in results:
            message += '  {}. "{}"\n'.format(rank, result[0])
            rank += 1
        logging.info(message)
        context.bot.send_message(chat_id, text=message, parse_mode='HTML', reply_id=None)


def ask(update, context):
    handle_query(update, context)


def image_handler(bot, update):
    file_id = update.message.photo.file_id
    chat_id = update.message.chat.id
    log.info('Received image. chat-id: {}. file-id: {}', chat_id, file_id)
    file = bot.get_file(file_id)
    image_uri = '/tmp/{}.jpg'.format(update.message.photo.file_id)
    log.info('Download image: "%s"', image_uri)
    file.download(image_uri)

    namespace = "chat_" + str(update.message.chat.id)
    results = engine.search(namespace,
                            3,
                            'file:{}'.format(image_uri)
                            , 'image')

    # For testing
    # results = [
    #     ("https://storage.googleapis.com/imagestash001/906802_3a8ac174-a9cc-4e46-88b3-bfda75941d9f_1076_1061.png", 0),
    #     ("https://storage.googleapis.com/imagestash001/23924271_8b9fb814-228b-4d4e-b51c-3e7943cd1c43_1080_1080.jpg", 1),
    #     ("https://storage.googleapis.com/imagestash001/122138742_71ddecc0-6531-4007-a622-91b076c0b602_1208_1208.jpg", 2)
    # ]
    for result in results:
        bot.send_photo(chat_id, result[0])


def process(update, context):
    if update.message == None:
        return

    if update.message.message_id == None:
        return

    if update.message.text == None:
        return

    embeddings = []
    if update.message.reply_to_message != None:
        if update.message.reply_to_message.text != None:
            embeddings.append(update.message.reply_to_message.text)
    embeddings.append(update.message.text)

    namespace = "chat_"+str(update.message.chat.id)
    reference = str(update.message.message_id)

    if engine.index(namespace, reference, embeddings):
        log.info("%s/%s, %s inserted", namespace, reference, embeddings)
    else:
        log.info("%s/%s, %s updated", namespace, reference, embeddings)


def process_entity(update, context):
    log.info(update)


def error(update, context):
    log.warning('"%s" error "%s"', update, context.error)


def serve(bot_token=None):
    if bot_token is None and "BOT_TOKEN" not in os.environ:
        log.error("Environment variable BOT_TOKEN not defined, bot won't start!")
        return None

    if os.environ.get("BOT_TOKEN") is not None:
        bot_token = os.environ["BOT_TOKEN"]
    updater = Updater(bot_token, use_context=True)

    dp = updater.dispatcher
    dp.add_handler(CommandHandler("start", start))
    dp.add_handler(CommandHandler("help", help))
    dp.add_handler(CommandHandler("ask", ask))
    dp.add_handler(MessageHandler(Filters.photo, image_handler))
    dp.add_handler(MessageHandler(Filters.text, process))
    dp.add_handler(MessageHandler(Filters.sticker, process_entity))
    dp.add_error_handler(error)

    print('Started bot service ...')
    updater.start_polling()
    return updater


def main():
    updater = serve()
    if updater is not None:
        updater.idle()


if __name__ == '__main__':
    main()



""" Generate proto with gRPC plugin to generate messages and gRPC stubs.

"""
from grpc_tools import protoc

protoc.main((
    '',
    '--proto_path=.',
    '--python_out=.',
    '--grpc_python_out=.',
    'hawking.proto',
))



""" Bert encoder (Online | Offline).

Online: bert text encoder using bert-as-a-service encoder. github: https://github.com/hanxiao/bert-as-service.
Offline: in process bert text encoder. github: https://github.com/huggingface/pytorch-transformers.
"""
import time
import numpy as np
import torch
from pytorch_transformers import BertModel, BertTokenizer
from bert_serving.client import BertClient
from hawking.encoder.universal import UniversalTextEncoder


class OnlineBertTextEncoder:
    def __init__(self, server_ip='127.0.0.1', port=5555):
        self.server_ip = server_ip
        self.client = BertClient(ip=self.server_ip, port=port)

    def async_encode(self, feeder, callback, context):
        for v in self.client.encode_async(feeder(), max_num_batch=context.size):
            callback(v, context)

    def batch_encode(self, phrases):
        """ Batch encode phrases into embedding vectors. """
        return self.client.encode(phrases)

    def encode(self, phrase):
        """ Encode single phrase into an embedding vector. """
        return self.batch_encode([phrase])


class BertTextEncoder:
    def __init__(self, pretrained_weights='bert-base-uncased'):
        model_class = BertModel
        tokenizer_class = BertTokenizer

        # Load pretrained model/tokenizer
        st = time.time()
        print('Loading BERT pre-trained model `{}`'.format(pretrained_weights))
        self.tokenizer = tokenizer_class.from_pretrained(pretrained_weights)
        self.model = model_class.from_pretrained(pretrained_weights)
        print('Loading BERT model completed. Elapsed time: %.2f secs' % (time.time() - st))

    def batch_encode(self, phrases):
        features = []
        for phrase in phrases:
            input_ids = torch.tensor([self.tokenizer.encode(phrase, add_special_tokens=True)])
            with torch.no_grad():
                # Models outputs are tuples
                # TODO(oleg): get second last layer
                last_hidden_states = self.model(input_ids)[0]
                features.append(np.average(last_hidden_states[0], axis=0))

        return features

    def encode(self, phrase):
        return self.batch_encode([phrase])[0]


def encode_phrase(phrase, server_ip='127.0.0.1', port=5555,
                  weights='universal-sentence-encoder-multilingual-qa/1'):
    """ single-line wrapper of XXXXBertTextEncoder. """
    if encode_phrase.enc is None:
        if weights.startswith('bert-base-'):
            encode_phrase.enc = BertTextEncoder(pretrained_weights=weights)
        elif weights == 'online-bert':
            encode_phrase.enc = OnlineBertTextEncoder(server_ip=server_ip, port=port)
        elif weights.startswith('universal-sentence-'):
            encode_phrase.enc = UniversalTextEncoder(module_url='https://tfhub.dev/google/' + weights)
        else:
            raise ValueError('Weight `{}` is not recognized.'.format(weights))
    return encode_phrase.enc.encode(phrase)


encode_phrase.enc = None



import logging
import sys


def configure_logging():
    """ Configure logging to stream to std.out """
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    # handler = logging.StreamHandler(sys.stdout)
    # handler.setLevel(logging.INFO)
    # formatter = logging.Formatter('%(asctime)s,%(name)s,%(levelname)s,%(message)s')
    # handler.setFormatter(formatter)
    # root.addHandler(handler)



""" ANSI Escape Color Code.

Define some ANSI escape code for shell output color.
"""
BLACK="0;30"
DARK_GRAY="1;30"
CYAN="0;36"
LIGHT_CYAN="0;96"
LIGHT_GRAY="0;37"
RED="0;31"
GREEN="0;32"
LIGHT_GREEN="0;92"
ORANGE="0;33"
YELLOW="1;33"
LIGHT_YELLOW="0;93"
BLUE="0;34"
LIGHT_BLUE="0:94"
PURPLE="0;35"
WHITE="1;37"
BRIGHT_WHITE="0;97"


def print_color(skk, color=WHITE, underline=False, bold=False):
    """ Color coded print out to stdout. """
    decoration = ';4' if underline else ''
    decoration = ';1' if bold else decoration
    print("\033[{}{}m {}\033[00m".format(color, decoration, skk))




""" Secretary Telegram Bot. """
import logging
import json
import os
import random
import sys

from telegram import Sticker
from telegram.ext import Updater, CommandHandler, MessageHandler, Filters

from hawking.frontend.engine.online import engine

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
log.addHandler(handler)



def start(update, context):
    log.info("Conversation started!")
    update.message.reply_text('Welcome, to next gen search engine')


def help(update, context):
    log.info("Help requested!")
    update.message.reply_text("I'm listening to conversation and taking notes. If you want to lookup details, send /q <term>")


def question(update, context):
    if update.message == None:
        return

    if update.message.text == None:
        return

    namespace = "secretary_"+str(update.message.chat.id)
    query = update.message.text
    if query.startswith("/q"):
        query = query[len("/q"):]
        query = query.strip()

    log.info('search within "%s" namespace for "%s"', namespace, query)

    results = engine.search(namespace, 2, query)

    chat_id = update.message.chat.id
    for result in results:
        context.bot.send_message(chat_id, text="Score: " + str(result[1]), parse_mode='MARKDOWN', reply_to_message_id=result[0])

def process(update, context):
    if update.message == None:
        return

    if update.message.message_id == None:
        return

    if update.message.text == None:
        return

    phrase = ""
    if update.message.reply_to_message != None:
        if update.message.reply_to_message.text != None:
            phrase += update.message.reply_to_message.text
    phrase += update.message.text

    namespace = "secretary_"+str(update.message.chat.id)
    reference = str(update.message.message_id)

    if engine.index(namespace, reference, phrase):
        log.info("%s/%s, %s inserted", namespace, reference, phrase)
    else:
        log.info("%s/%s, %s updated", namespace, reference, phrase)


def process_entity(update, context):
    log.info(update)


def error(update, context):
    log.warning('"%s" error "%s"', update, context.error)


def serve(bot_token=None):
    if bot_token is None and "SECRETARY_BOT_TOKEN" not in os.environ:
        log.error("Environment variable SECRETARY_BOT_TOKEN not defined, bot won't start!")
        return None

    if os.environ.get("SECRETARY_BOT_TOKEN") is not None:
        bot_token = os.environ["SECRETARY_BOT_TOKEN"]
    updater = Updater(bot_token, use_context=True)

    dp = updater.dispatcher
    dp.add_handler(CommandHandler("start", start))
    dp.add_handler(CommandHandler("help", help))
    dp.add_handler(CommandHandler("q", question))
    dp.add_handler(MessageHandler(Filters.text, process))
    dp.add_handler(MessageHandler(Filters.sticker, process_entity))
    dp.add_error_handler(error)

    print('Started bot service ...')
    updater.start_polling()
    return updater


def main():
    updater = serve()
    if updater is not None:
        updater.idle()


if __name__ == '__main__':
    main()



import logging

from flask import request
from flask_restplus import Api, Resource, reqparse, fields
from hawking.frontend.engine.default import engine as default_engine
from hawking.frontend.engine.online import engine as online_engine
from hawking.frontend.store.default import store

log = logging.getLogger(__name__)

api = Api(version='1.0', title='Hawking API', description='Hawking powered search engine')
ns = api.namespace('hawking', description='Search engine operation')

search_arguments = reqparse.RequestParser()
search_arguments.add_argument('namespace', type=str, required=True, default=0, help='Namespace/scope')
search_arguments.add_argument('query', type=str, required=True, default=0, help='Query string')
search_arguments.add_argument('offset', type=int, required=False, default=0, help='Page offset')
search_arguments.add_argument('count', type=int, required=False, choices=[3, 10, 15, 20],
                                default=10, help='Results per page {error_msg}')

result = api.model('Result', {
    'uuid': fields.String(required=True, description='Result unique reference'),
    'score': fields.Float(required=True, description='Result score'),
})

page_of_results = api.model('Page of results', {
    'results': fields.List(fields.Nested(result)),
})

def search_records(request, engine):
    """
    Search.
    """
    args = search_arguments.parse_args(request)

    namespace = args.get('namespace')
    query = args.get('query')
    result_offset = args.get('offset', 0)
    result_count = args.get('count', 10)

    log.info('%s: search within "%s" namespace for "%s", look for %d results starting with %d offset',
                type(engine), namespace, query, result_count, result_offset)

    results = engine.search(namespace, result_offset + result_count, query)
    log.info('pre %s', results)

    results = results[result_offset:]
    log.info('post %s', results)

    responses = []
    for result in results:
        responses.append({"uuid": result[0], "score": result[1]})

    log.info('resp %s', responses)
    return {"results": responses}


@ns.route('/search')
class DefaultSearchApi(Resource):
    @api.expect(search_arguments)
    @api.marshal_with(page_of_results)
    def get(self):
        return search_records(request, default_engine)
        

@ns.route('/search/online')
class OnlineSearchApi(Resource):
    @api.expect(search_arguments)
    @api.marshal_with(page_of_results)
    def get(self):
        return search_records(request, online_engine)


record = api.model('Record', {
    'uuid': fields.String(required=True, description='Record unique reference'),
    'namespace': fields.String(required=True, description='Record namespace'),
    'content': fields.String(required=True, description='Record content'),
})

def index_record(request, engine):
    """
    Index record.
    """
    args = request.json
    namespace = args.get('namespace')
    uuid = args.get('uuid')
    content = args.get('content')

    if engine.index(namespace, uuid, content):
        log.info("%s/%s/%s: %s inserted", type(engine), namespace, uuid, content)
    else:
        log.info("%s/%s/%s: %s updated", type(engine), namespace, uuid, content)

    return None, 201

@ns.route('/index')
class IndexApi(Resource):
    @api.expect(record)
    def put(self):
        return index_record(request, default_engine)
        

@ns.route('/index/online')
class OnlineIndexApi(Resource):
    @api.expect(record)
    def put(self):
        return index_record(request, online_engine)



store_put_result = api.model('UUID of stored documents', {
    'uuid': fields.String(),
})


store_get = reqparse.RequestParser()
store_get.add_argument('namespace', type=str, required=True, default=0, help='Namespace/scope')
store_get.add_argument('uuid', type=str, required=True, default=0, help='Document unique reference')


store_get_result = api.model('Document', {
    'uuid': fields.String(),
    'content': fields.String(),
})

@ns.route('/store')
class OnlineStoreApi(Resource):
    @api.expect(record)
    @api.marshal_with(store_put_result)
    def put(self):
        """
        Store object.
        """
        args = request.json
        namespace = args.get('namespace')
        uuid = args.get('uuid')
        content = args.get('content')

        uuid = store.put(namespace, uuid, content)
        log.info("Store %s/%s: %s", namespace, uuid, content)

        return {"uuid": uuid}, 201

    @api.expect(store_get)
    @api.marshal_with(store_get_result)
    def get(self):
        """
        Search.
        """
        args = store_get.parse_args(request)

        namespace = args.get('namespace')
        uuid = args.get('uuid')

        log.info("Retrieve %s/%s", namespace, uuid)

        content = store.get(namespace, uuid)
        return {"uuid": uuid, "content": content}


@api.errorhandler(ValueError)
def value_error_handler(e):
    message = str(e)
    log.exception(message)

    return {'message': message}, 401


@api.errorhandler
def default_error_handler(e):
    message = 'An unhandled exception occurred.'
    log.exception(message)

    return {'message': message}, 500




""" Faiss Index Server.

"""
import os
import json
import logging
import time

import click
import faiss
import uuid
import grpc
import numpy as np

from concurrent import futures
from urllib.request import urlretrieve
from hawking_proto import hawking_pb2, hawking_pb2_grpc
from hawking.encoder import create_encoder
from google.cloud import storage
from hawking.serve.caption_server import CaptionServicer
from hawking.serve import configure_logging


_ONE_DAY_IN_SECONDS = 60 * 60 * 24


def download_gcs_blob(bucket_name, source_blob_name, destination_file_name):
    """Downloads a blob from the bucket."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(source_blob_name)

    blob.download_to_filename(destination_file_name)

    print('Blob {} downloaded to {}.'.format(
        source_blob_name,
        destination_file_name))


class SearchServicer(hawking_pb2_grpc.SearchServicer):
    """ Implements hawking_pb2_grpc.SearchServicer. """

    def __init__(self, index_path, indexer_host, indexer_port, caption_host, caption_port, encoder_config):
        del indexer_host
        del indexer_port

        self.ready = False
        self.caption_stub = None
        self.caption_servicer = None

        if caption_host == '127.0.0.1':
            logging.info('Loading caption service ... {}:{}'.format(caption_host, caption_port))
            filename, ext = os.path.splitext(index_path)
            self.caption_servicer = CaptionServicer('{}.caption'.format(filename))
        else:
            logging.info('Connecting to caption service ... {}:{}'.format(caption_host, caption_port))
            self.channel = grpc.insecure_channel('{}:{}'.format(caption_host, caption_port))
            self.caption_stub = hawking_pb2_grpc.CaptionStub(self.channel)

        self.LoadIndex(index_path)
        logging.info(encoder_config)
        self.entity_type = encoder_config['entity_type']
        self.encoder = create_encoder(config=encoder_config)
        logging.info('Fx is ready to serve ...')
        self.ready = True

    def GetCaption(self, request):
        return (self.caption_stub.Get(request)
                if self.caption_stub is not None
                else self.caption_servicer.Get(request, None))

    def LoadIndex(self, index_path):
        assert os.path.exists(index_path), "Index {} does not exist!".format(index_path)
        logging.info('Loading index {}'.format(index_path))
        st = time.time()
        self.index = faiss.read_index(index_path)
        et = time.time() - st
        logging.info('Index loaded. Elapsed time: %.2f s' % et)

    def Get(self, request, context):
        logging.info('Receive query: {}'.format(request))
        if not self.ready:
            raise Exception('Server is not ready to serve ...')
        query_response = hawking_pb2.SearchResponse()

        try:
            logging.info(self.entity_type)
            if request.type == hawking_pb2.SearchRequest.EntityType.IMAGE:
                image_uri = request.query
                if image_uri.startswith('file:'):
                    tmp_file = image_uri[len('file:'):]
                else:
                    tmp_file = '/tmp/' + str(uuid.uuid4())
                    logging.info('Downloading {} to {}'.format(image_uri, tmp_file))
                    if image_uri.startswith('http'):
                        urlretrieve(image_uri, tmp_file)
                    elif image_uri.startswith('gs://'):
                        tmp_uri = image_uri
                        tokens = tmp_uri[:6].split('/')
                        bucket_name = tokens[0]
                        source_uri = '/'.join(tokens[:1])
                        download_gcs_blob(bucket_name, source_uri, tmp_file)
                # TODO(oleg): do base64 format.
                vector = self.encoder.encode(tmp_file, flattened=False)
            elif request.type == hawking_pb2.SearchRequest.EntityType.TEXT:
                vector = self.encoder.encode(request.query)
            else:
                raise ValueError('Query for entity type {} is not supported.'.format(
                    request.type))
            np.set_printoptions(threshold=5)
            logging.info('Obtained {} vector({}): {}'.format(request.query, vector.shape, vector))
            logging.info('Querying index...')
            D, I = self.index.search(vector, request.max_results)

            # Retrieve captions from caption_server, for image type it returns image uri,
            # for text, it will return metadata.
            logging.info('Querying caption...')
            caption_request = hawking_pb2.SnippetRequest(max_length=250, prettify=False)
            for doc_id in I[0]:
                caption_request.doc_ids.append(doc_id)

            st = time.time()
            caption_response = self.GetCaption(caption_request)

            logging.info('Search results:')
            for snippet, doc_id, score in zip(caption_response.snippets, I[0], D[0]):
                query_response.results.append(
                    hawking_pb2.Result(
                        score=score,
                        document=hawking_pb2.Document(uuid=str(doc_id), body=snippet.title)))

            et = (time.time() - st) * 1000.0
            logging.info('elapsed time: %d ms' % et)
            logging.info('Response:{}'.format(query_response))
        except Exception as e:
            logging.error(e)
        return query_response


def serve(index_path, indexer_host, indexer_port, caption_host, caption_port, worker_threads, grpc_port,
          encoder_config, run_bot):
    configure_logging()

    logging.info('TAG: FX')
    logging.info('Starting gRPC index server ...')
    logging.info(encoder_config)
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=worker_threads))
    search_servicer = SearchServicer(
        index_path,
        indexer_host, indexer_port,
        caption_host, caption_port,
        encoder_config)
    hawking_pb2_grpc.add_SearchServicer_to_server(search_servicer, server)
    server.add_insecure_port('[::]:{}'.format(grpc_port))
    server.start()
    #
    # from hawking.bot.secretary import serve as start_secretary
    # from hawking.bot.telegram import serve as start_bot
    # bot = None
    # secretary = None
    # if run_bot:
    #     bot = start_bot(bot_token='701884073:AAF9rkmy73sq6U3QPbZxX5JzyfkDqhP-5xQ')
    #     # TODO(oleg): add secretary bot token.
    #     secretary = None

    logging.info('Ready to serve. Listening on 0.0.0.0:{}'.format(grpc_port))
    logging.info('Press Ctrl+C to exit ...')


    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        # if bot is not None:
        #    bot.stop()
        # if secretary is not None:
        #    secretary.stop()
        if server is not None:
            server.stop(0)


@click.command()
@click.option('--index-path', help="Index path.", default=None)
@click.option('--indexer-host', help="Indexer backend host url/ip.", default='127.0.0.1')
@click.option('--indexer-port', help="Indexer backend host port.", default=5000)
@click.option('--caption-host', help="Caption server host url/ip.", default='127.0.0.1')
@click.option('--caption-port', help="Caption server port.", default=50001)
@click.option('--worker-threads', '-w', help="Maximum numbers of worker threads.", default=10)
@click.option('--grpc-port', help="gRPC port to listen to request.", default=50000)
@click.option('--encoder-config', help="Encoder pre-trained weights.", default='/tmp/encoder_config.json')
@click.option('--run-bot/--no-run-bot', help="Run telegram bot.", default=False)
def run(index_path, indexer_host, indexer_port, caption_host, caption_port, worker_threads,
        grpc_port, encoder_config, run_bot):
    """Run server
    example of encoder_config:
    {
        'entity_type': 'text',
        'encoder_type': 'universal-sentence-encoder',
        'weights': 'https://tfhub.dev/google/universal-sentence-encoder-multilingual-qa/1',
        'encoder_host': '127.0.0.1',
        'encoder_port': 5555,
        'top_layer': "fc1000",
        'print_summary': False,
    }
    """
    with open(encoder_config, 'r') as f:
        encoder_config = json.load(f)
    serve(index_path,
          indexer_host, int(indexer_port),
          caption_host, int(caption_port),
          int(worker_threads), int(grpc_port), encoder_config, run_bot)


if __name__ == "__main__":
    run()



import os
import click

from hawking.encoder.vgg import VGGImageEncoder
from hawking.encoder.resnet50 import ResNet50ImageEncoder


@click.command()
@click.argument('image_dir')
@click.option('--arch', '-a', help="Underlying Neural Nets architecture: vgg16, resnet50", default="resnet50")
def encode_images(image_dir, arch):
    files = os.listdir(image_dir)
    if arch == 'vgg16':
        enc = VGGImageEncoder()
    elif arch == 'resnet50':
        enc = ResNet50ImageEncoder()
    for f in files:
        path = os.path.join(image_dir, f)
        v = enc.encode(path)
        print(v.shape)
        print(v)
        input()


if __name__ == '__main__':
    encode_images()



""" Resnet50 image encoder.

Returns the convolutional features of Resnet50.
"""
import logging
import numpy as np
import tensorflow as tf

from keras.preprocessing import image
from keras.applications.resnet50 import ResNet50
from keras.applications.resnet50 import preprocess_input
from keras.models import Model


class ResNet50ImageEncoder:
    def __init__(self, print_summary=True, top_layer='fc1000', include_top=True):
        base_model = ResNet50(weights='imagenet', include_top=include_top)
        self.model = Model(inputs=base_model.input, outputs=base_model.get_layer(top_layer).output)
        self.graph = tf.get_default_graph()
        if print_summary:
            self.model.summary()

    def batch_encode(self, img_paths):
        features = []
        for img_path in img_paths:
            features.append(self.encode(img_path))
        return features

    def encode(self, img_path, flattened=True):
        # Running predict on webservice, requires fetching tf.graph
        # See: https://github.com/keras-team/keras/issues/2397
        with self.graph.as_default():
            logging.info('Encode: {}'.format(img_path))
            img = image.load_img(img_path, target_size=(224, 224))
            img_data = image.img_to_array(img)
            img_data = np.expand_dims(img_data, axis=0)
            img_data = preprocess_input(img_data)
            feature = self.model.predict(img_data)
            feature_np = np.array(feature)
            return feature_np.flatten() if flattened else feature_np



""" Queue Server.

"""
import os
import json
import logging
import sqlite3
import sys
import time
import threading

import click
import grpc

from concurrent import futures
from hawking_proto import hawking_pb2, hawking_pb2_grpc
from hawking.encoder.bert import BertTextEncoder


_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class QueueServicer(hawking_pb2_grpc.QueueServicer, hawking_pb2_grpc.IndexServicer):
    """ Implements hawking_pb2_grpc.QueueServicer. """

    def __init__(self, queue_path):
        self.db_file = os.path.join(queue_path, "queue.db")
        logging.info("Use {} file".format(self.db_file))

        connection = sqlite3.connect(self.db_file)
        try:
            connection.execute("CREATE TABLE IF NOT EXISTS queue (checkpoint INTEGER PRIMARY KEY, namespace VARCHAR(1024), document JSON NOT NULL);")
            connection.commit()
        finally:
            connection.close()

        self.condition = threading.Condition()


    def _poll(self, checkpoint):
        connection = sqlite3.connect(self.db_file)
        try:
            cursor = connection.cursor()
            cursor.execute("SELECT namespace, document FROM queue WHERE checkpoint = ?", (checkpoint,) )

            row = cursor.fetchone()
            if row is None:
                return None

            return {"namespace": row[0], "document": row[1] }
        finally:
            connection.close()

    def Poll(self, request, context):
        """
            option (google.api.http) = {
                post: "/v1/post"
                body: "*"
            };
        """
        checkpoint = 1 if request.checkpoint is None or request.checkpoint == '' else (int(request.checkpoint))
        logging.info("{} start with checkpoint {}".format(context.peer(), checkpoint))

        while True:
            response = None
            self.condition.acquire()

            item = self._poll(checkpoint)
            while item is None:
                self.condition.wait()
                item = self._poll(checkpoint)

            checkpoint += 1

            namespace = item['namespace']
            record = json.loads(item['document'])
            response = hawking_pb2.PollResponse(checkpoint = str(checkpoint),
                                                namespace = namespace,
                                                document = hawking_pb2.Document(uuid=record['uuid'], body=record['content']))

            self.condition.release()
            yield response


        


    """ Implements hawking_pb2_grpc.IndexServicer. """

    def Post(self, request, context):
        """
            option (google.api.http) = {
                post: "/v1/post"
                body: "*"
            };
        """
        self.condition.acquire()

        connection = sqlite3.connect(self.db_file)
        try:
            for document in request.documents:
                logging.info('Enqueue "{}/{}/{}"'.format(request.namespace, document.uuid, document.body))
                connection.execute("INSERT INTO queue (namespace, document) VALUES (?, ?)",
                    (request.namespace, json.dumps({"uuid":document.uuid, "content": document.body}))
                )

            connection.commit()
        finally:
            connection.close()

        self.condition.notifyAll()
        self.condition.release()

        return hawking_pb2.IndexResponse(inserted=True)



def serve(queue_path, worker_threads, grpc_port):
    logging.info('TAG: Queue')
    logging.info('Starting gRPC queue server ...')
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=worker_threads))
    queue_servicer = QueueServicer(queue_path)
    hawking_pb2_grpc.add_QueueServicer_to_server(queue_servicer, server)
    hawking_pb2_grpc.add_IndexServicer_to_server(queue_servicer, server)
    server.add_insecure_port('[::]:{}'.format(grpc_port))
    server.start()

    logging.info('Ready to serve. Listening on 0.0.0.0:{}'.format(grpc_port))
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


@click.command()
@click.option('--queue-path', help="Queue dir for stuff.", default="/tmp")
@click.option('--worker-threads', '-w', help="Maximum numbers of worker threads.", default=10)
@click.option('--grpc-port', help="gRPC port to listen to request.", default=45000)
def run(queue_path, worker_threads, grpc_port):
    serve(queue_path, int(worker_threads), int(grpc_port))


if __name__ == "__main__":
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    run()



""" Node Server.

"""
import os
import logging
import sys

import time
import threading

import click
import grpc
import sqlite3

from concurrent import futures
from hawking_proto import hawking_pb2, hawking_pb2_grpc



_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class StoreServicer(hawking_pb2_grpc.StoreServicer):
    """ Implements hawking_pb2_grpc.StoreServicer. """

    def __init__(self, store_path):
        self.namespaces = {}
        self.db_file = os.path.join(store_path, "store.db")
        logging.info("Use {} file".format(self.db_file))
        connection = sqlite3.connect(self.db_file)
        try:
            connection.execute("CREATE TABLE IF NOT EXISTS data (namespace VARCHAR(1024), uuid VARCHAR(1024), content JSON NOT NULL, PRIMARY KEY (namespace, uuid));")
            connection.commit()
        finally:
            connection.close()

    def Put(self, request, context):
        connection = sqlite3.connect(self.db_file)
        try:
            connection.execute("INSERT OR REPLACE INTO data (namespace, uuid, content) VALUES (?, ?, ?)",
                (request.namespace, request.document.uuid, request.document.body)
            )
            connection.commit()
        finally:
            connection.close()

        logging.info("Store {}/{}: {}".format(request.namespace, request.document.uuid, request.document.body))
        return hawking_pb2.StorePutResponse(uuid=request.document.uuid)


    def Get(self, request, context):
        connection = sqlite3.connect(self.db_file)
        try:
            cursor = connection.cursor()
            cursor.execute("SELECT content FROM data WHERE namespace = ? AND uuid = ?",
                (request.namespace, request.uuid)
            )

            row = cursor.fetchone()
            if row is None:
                context.set_code(grpc.StatusCode.NOT_FOUND)
                context.set_details("Object `" + request.namespace + "/" + request.uuid + "` not found!")
                raise ValueError("Object `" + request.namespace + "/" + request.uuid + "` not found!")

            content = row[0]
            logging.info("Retrieve {}/{}: {}".format(request.namespace, request.uuid, content))
            return hawking_pb2.StoreGetResponse(document=
                hawking_pb2.Document(body=content, uuid=request.uuid))
        finally:
            connection.close()


def serve(store_path, worker_threads, grpc_port):
    logging.info('TAG: Store')
    logging.info('Starting gRPC queue server ...')
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=worker_threads))
    queue_servicer = StoreServicer(store_path)
    hawking_pb2_grpc.add_StoreServicer_to_server(queue_servicer, server)
    server.add_insecure_port('[::]:{}'.format(grpc_port))
    server.start()

    logging.info('Ready to serve. Listening on 0.0.0.0:{}'.format(grpc_port))
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


@click.command()
@click.option('--store-path', help="Store dir for stuff.", default="/tmp")
@click.option('--worker-threads', '-w', help="Maximum numbers of worker threads.", default=10)
@click.option('--grpc-port', help="gRPC port to listen to request.", default=42000)
def run(store_path, worker_threads, grpc_port):
    serve(store_path, int(worker_threads), int(grpc_port))


if __name__ == "__main__":
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    run()



""" Faiss Index Server.

"""
import os
import logging
import sys
import time

import click
import grpc
import numpy as np

from concurrent import futures
from hawking_proto import hawking_pb2, hawking_pb2_grpc
from hawking.encoder.bert import BertTextEncoder
from hawking.sptag import SPTAGClient


_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class SearchServicer(hawking_pb2_grpc.SearchServicer):
    """ Implements hawking_pb2_grpc.SearchServicer. """

    def __init__(self, index_path, indexer_host, indexer_port, encoder_host, encoder_port):
        self.LoadIndex(index_path)
        logging.info('Connecting to text encoder service `{}:{}`...'.format(encoder_host, encoder_port))
        self.encoder = BertTextEncoder(encoder_host, encoder_port)
        logging.info('Successfully connected to text encoder service')

        logging.info('Connecting to indexer backend `{}:{}` ...'.format(indexer_host, indexer_port))
        self.sptag_client = SPTAGClient.AnnClient(indexer_host, str(indexer_port))
        while not self.sptag_client.IsConnected():
            logging.info('Connecting ...')
            time.sleep(1)
        logging.info('Successfully connected to indexer backend ...')

    def LoadIndex(self, index_path):
        pass

    def Get(self, request, context):
        logging.info('Receive query: {}'.format(request))
        st = time.time()
        vector = self.encoder.encode(request.query)
        result = self.sptag_client.Search(vector, request.max_results, 'Float', True)
        response = hawking_pb2.SearchResponse()
        for docid, dist, meta in zip(result[0], result[1], result[2]):
            response.results.append(
                    hawking_pb2.Result(
                        score=dist,
                        document=hawking_pb2.Document(uuid=str(docid), body=meta, parent=str(docid))))
        et = time.time() - st
        logging.info('Elapsed time: %d ms' % (et * 1000))
        return response


def serve(index_path, indexer_host, indexer_port, encoder_host, encoder_port, worker_threads, grpc_port):
    logging.info('TAG: SX')
    logging.info('Starting gRPC index server ...')
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=worker_threads))
    search_servicer = SearchServicer(index_path, indexer_host, indexer_port, encoder_host, encoder_port)
    hawking_pb2_grpc.add_SearchServicer_to_server(search_servicer, server)
    server.add_insecure_port('[::]:{}'.format(grpc_port))
    server.start()
    logging.info('Ready to serve. Listening on 0.0.0.0:{}'.format(grpc_port))
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


@click.command()
@click.option('--index-path', '-idx', help="Index path.", default=None)
@click.option('--indexer-host', help="Indexer backend host url/ip.", default='127.0.0.1')
@click.option('--indexer-port', help="Indexer backend host port.", default=5000)
@click.option('--encoder-host', help="Remote text encoder host url/ip.", default='127.0.0.1')
@click.option('--encoder-port', help="Remote text encoder host port.", default=5555)
@click.option('--worker-threads', '-w', help="Maximum numbers of worker threads.", default=10)
@click.option('--grpc-port', help="gRPC port to listen to request.", default=50000)
def run(index_path, indexer_host, indexer_port, encoder_host, encoder_port, worker_threads, grpc_port):
    serve(index_path, indexer_host, indexer_port, encoder_host, encoder_port, int(worker_threads), grpc_port)


if __name__ == "__main__":
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    run()



import logging
import time
import grpc

from hawking_proto import hawking_pb2, hawking_pb2_grpc

# Default dummy store
class DummyStore():
    def __init__(self):
        self.namespaces = {}
        self.store_host = '127.0.0.1'
        self.store_port = 42000
        self.channel = grpc.insecure_channel('{}:{}'.format(self.store_host, self.store_port))

    # store object (content_type, blob) under namespace/uuid
    # returns: uuid
    def put(self, namespace, uuid, content):
        with grpc.insecure_channel('{}:{}'.format(self.store_host, self.store_port)) as channel:
            stub = hawking_pb2_grpc.StoreStub(channel)
            logging.info('Put: "{}/{}"'.format(namespace, uuid))
            request = hawking_pb2.StorePutRequest(namespace = namespace, document=hawking_pb2.Document(uuid = uuid, body=content))
            
            return stub.Put(request).uuid

    # retrieve object (content_type, blob) by namespace/uuid
    #
    # returns: {content_type, blob}
    def get(self, namespace, uuid):
        with grpc.insecure_channel('{}:{}'.format(self.store_host, self.store_port)) as channel:
            stub = hawking_pb2_grpc.StoreStub(channel)
            logging.info('Get: "{}/{}"'.format(namespace, uuid))
            request = hawking_pb2.StoreGetRequest(namespace = namespace, uuid = uuid)
            return stub.Get(request).document.body

store = DummyStore()



import SPTAG
import numpy as np

n = 100
k = 3
r = 3

def testBuild(algo, distmethod, x, out):
   i = SPTAG.AnnIndex(algo, 'Float', x.shape[1])
   i.SetBuildParam("NumberOfThreads", '4')
   i.SetBuildParam("DistCalcMethod", distmethod)
   ret = i.Build(x, x.shape[0])
   i.Save(out)

def testBuildWithMetaData(algo, distmethod, x, s, out):
   i = SPTAG.AnnIndex(algo, 'Float', x.shape[1])
   i.SetBuildParam("NumberOfThreads", '4')
   i.SetBuildParam("DistCalcMethod", distmethod)
   if i.BuildWithMetaData(x, s, x.shape[0]):
       i.Save(out)

def testSearch(index, q, k):
   j = SPTAG.AnnIndex.Load(index)
   for t in range(q.shape[0]):
       result = j.Search(q[t], k)
       print (result[0]) # ids
       print (result[1]) # distances

def testSearchWithMetaData(index, q, k):
   j = SPTAG.AnnIndex.Load(index)
   j.SetSearchParam("MaxCheck", '1024')
   for t in range(q.shape[0]):
       result = j.SearchWithMetaData(q[t], k)
       print (result[0]) # ids
       print (result[1]) # distances
       print (result[2]) # metadata

def testAdd(index, x, out, algo, distmethod):
   if index != None:
       i = SPTAG.AnnIndex.Load(index)
   else:
       i = SPTAG.AnnIndex(algo, 'Float', x.shape[1])
   i.SetBuildParam("NumberOfThreads", '4')
   i.SetBuildParam("DistCalcMethod", distmethod)
   if i.Add(x, x.shape[0]):
       i.Save(out)

def testAddWithMetaData(index, x, s, out, algo, distmethod):
   if index != None:
       i = SPTAG.AnnIndex.Load(index)
   else:
       i = SPTAG.AnnIndex(algo, 'Float', x.shape[1])
   i = SPTAG.AnnIndex(algo, 'Float', x.shape[1])
   i.SetBuildParam("NumberOfThreads", '4')
   i.SetBuildParam("DistCalcMethod", distmethod)
   if i.AddWithMetaData(x, s, x.shape[0]):
       i.Save(out)

def testDelete(index, x, out):
   i = SPTAG.AnnIndex.Load(index)
   ret = i.Delete(x, x.shape[0])
   print (ret)
   i.Save(out)
   
def Test(algo, distmethod):
   x = np.ones((n, 10), dtype=np.float32) * np.reshape(np.arange(n, dtype=np.float32), (n, 1))
   q = np.ones((r, 10), dtype=np.float32) * np.reshape(np.arange(r, dtype=np.float32), (r, 1)) * 2
   m = ''
   for i in range(n):
       m += str(i) + '\n'

   m = m.encode()

   print ("Build.............................")
   testBuild(algo, distmethod, x, 'testindices')
   testSearch('testindices', q, k)
   print ("Add.............................")
   testAdd('testindices', x, 'testindices', algo, distmethod)
   testSearch('testindices', q, k)
   print ("Delete.............................")
   testDelete('testindices', q, 'testindices')
   testSearch('testindices', q, k)

   print ("AddWithMetaData.............................")
   testAddWithMetaData(None, x, m, 'testindices', algo, distmethod)
   print ("Delete.............................")
   testSearchWithMetaData('testindices', q, k)
   testDelete('testindices', q, 'testindices')
   testSearchWithMetaData('testindices', q, k)

if __name__ == '__main__':
   Test('BKT', 'L2')
   Test('KDT', 'L2')



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.12
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info >= (2, 7, 0):
    def swig_import_helper():
        import importlib
        pkg = __name__.rpartition('.')[0]
        mname = '.'.join((pkg, '_SPTAG')).lstrip('.')
        try:
            return importlib.import_module(mname)
        except ImportError:
            return importlib.import_module('_SPTAG')
    _SPTAG = swig_import_helper()
    del swig_import_helper
elif _swig_python_version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_SPTAG', [dirname(__file__)])
        except ImportError:
            import _SPTAG
            return _SPTAG
        try:
            _mod = imp.load_module('_SPTAG', fp, pathname, description)
        finally:
            if fp is not None:
                fp.close()
        return _mod
    _SPTAG = swig_import_helper()
    del swig_import_helper
else:
    import _SPTAG
del _swig_python_version_info

try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr(self, class_type, name):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    raise AttributeError("'%s' object has no attribute '%s'" % (class_type.__name__, name))


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except __builtin__.Exception:
    class _object:
        pass
    _newclass = 0

SHARED_PTR_DISOWN = _SPTAG.SHARED_PTR_DISOWN
class AnnIndex(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, AnnIndex, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, AnnIndex, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _SPTAG.new_AnnIndex(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _SPTAG.delete_AnnIndex
    __del__ = lambda self: None

    def SetBuildParam(self, p_name, p_value):
        return _SPTAG.AnnIndex_SetBuildParam(self, p_name, p_value)

    def SetSearchParam(self, p_name, p_value):
        return _SPTAG.AnnIndex_SetSearchParam(self, p_name, p_value)

    def Build(self, p_data, p_num):
        return _SPTAG.AnnIndex_Build(self, p_data, p_num)

    def BuildWithMetaData(self, p_data, p_meta, p_num):
        return _SPTAG.AnnIndex_BuildWithMetaData(self, p_data, p_meta, p_num)

    def Search(self, p_data, p_resultNum):
        return _SPTAG.AnnIndex_Search(self, p_data, p_resultNum)

    def SearchWithMetaData(self, p_data, p_resultNum):
        return _SPTAG.AnnIndex_SearchWithMetaData(self, p_data, p_resultNum)

    def ReadyToServe(self):
        return _SPTAG.AnnIndex_ReadyToServe(self)

    def Save(self, p_saveFile):
        return _SPTAG.AnnIndex_Save(self, p_saveFile)

    def Add(self, p_data, p_num):
        return _SPTAG.AnnIndex_Add(self, p_data, p_num)

    def AddWithMetaData(self, p_data, p_meta, p_num):
        return _SPTAG.AnnIndex_AddWithMetaData(self, p_data, p_meta, p_num)

    def Delete(self, p_data, p_num):
        return _SPTAG.AnnIndex_Delete(self, p_data, p_num)
    if _newclass:
        Load = staticmethod(_SPTAG.AnnIndex_Load)
    else:
        Load = _SPTAG.AnnIndex_Load
AnnIndex_swigregister = _SPTAG.AnnIndex_swigregister
AnnIndex_swigregister(AnnIndex)

def AnnIndex_Load(p_loaderFile):
    return _SPTAG.AnnIndex_Load(p_loaderFile)
AnnIndex_Load = _SPTAG.AnnIndex_Load

# This file is compatible with both classic and new-style classes.





""" Google universal text encoder.

tf-hub-url: https://tfhub.dev/google/universal-sentence-encoder/2.
"""
import os
import time
import tensorflow as tf
import tensorflow_hub as hub
import tf_sentencepiece  # Not used directly but needed to import TF ops.


class UniversalTextEncoder:
    """ Google Universal Text Encoder. """
    def __init__(self,
            module_url='https://tfhub.dev/google/universal-sentence-encoder-large/3',
            module_cache_dir='/tmp/universal-encoder'):

        if not os.path.exists(module_cache_dir):
            os.system('mkdir -p {}'.format(module_cache_dir))
        st = time.time()
        os.environ['TFHUB_CACHE_DIR'] = module_cache_dir
        self.module_url = module_url
        print('Initializing tf....')
        g = tf.Graph()
        with g.as_default():
            self.text_input = tf.placeholder(dtype=tf.string, shape=[None])
            self.encoder = hub.Module(self.module_url)
            self.my_result = self.encoder(self.text_input)
            init_op = tf.group([tf.compat.v1.global_variables_initializer(), tf.compat.v1.tables_initializer()])
        g.finalize()    
        self.session = tf.compat.v1.Session(graph=g)
        self.session.run(init_op)

        print('Completed loading and initializing module: "%s". Elapsed time: "%.2f" s' % (self.module_url, time.time() - st))

    def batch_encode(self, phrases):
        """ Batch encode phrases into embedding vectors. """
        return self.session.run(self.my_result, feed_dict={self.text_input: phrases})

    def encode(self, phrase):
        """ Encode a single phrase into an embedding vector. """
        return self.batch_encode([phrase])





""" Hawking FrontEnd Server.

Copyright (C) 2019 Peach Inc. All rights reserved.
"""
import os
import logging.config

import click

logging_conf_path = os.path.normpath(os.path.join(os.path.dirname(__file__), 'logging.conf'))
logging.config.fileConfig(logging_conf_path)

from flask import Flask, Blueprint
from hawking.frontend.api.search import ns as engine_namespace
from hawking.frontend.api.search import api

blueprint = Blueprint('api', __name__, url_prefix='/api')
api.init_app(blueprint)
api.add_namespace(engine_namespace)

application = Flask(__name__)
application.config['SWAGGER_UI_DOC_EXPANSION'] = 'list'
application.config['RESTPLUS_VALIDATE'] = True
application.config['RESTPLUS_MASK_SWAGGER'] = False
application.config['ERROR_404_HELP'] = False
application.register_blueprint(blueprint)


def main(port):
    application.run(host='0.0.0.0', port=port)


@click.command()
@click.option('--port', '-p', help="Listening port", default=8080)
def run(port):
    main(port)


if __name__ == "__main__":
    main(8080)



import logging
import time
import grpc
import threading
import json
import numpy
import sys

from hawking_proto import hawking_pb2, hawking_pb2_grpc
from hawking.encoder.universal.encoder import UniversalTextEncoder

def _cos_similarity(f1, f2):
    n1 = numpy.linalg.norm(f1)
    n2 = numpy.linalg.norm(f2)

    return numpy.dot(f1, f2) / (n1 * n2)

class OnlineIndex():
    def __init__(self):
        self.encoder = UniversalTextEncoder()
        self.namespaces = {}
        self.checkpoint = None
        self.lock = threading.Lock()

        # TODO: move this to queue client
        self.queue_host = '127.0.0.1'
        self.queue_port = 45000
        self.retry_delay_sec = 30
        self.channel = grpc.insecure_channel('{}:{}'.format(self.queue_host, self.queue_port))
        logging.info("Start new OnlineSearchEngine")
        threading.Thread(daemon=True, name="poll-thread", target=self._poll).start()

    def _poll(self):
        stub = hawking_pb2_grpc.QueueStub(self.channel)
        while True:
            try:
                request = hawking_pb2.PollRequest(checkpoint = self.checkpoint)
                for index in stub.Poll(request):
                    logging.info('Checkpoint {} "{}/{}/{}" to index'.format(index.checkpoint, index.namespace, index.document.uuid, index.document.body))
                    features = self.encoder.encode(index.document.body)
                    self.checkpoint = index.checkpoint

                    with self.lock:
                        if index.namespace not in self.namespaces:
                            self.namespaces[index.namespace] = {}
                        self.namespaces[index.namespace][index.document.uuid] = features
            except grpc.RpcError as ex:
                logging.error(str(ex))
                logging.warn("Wait for {} seconds, before retry".format(self.retry_delay_sec))
                time.sleep(self.retry_delay_sec)
 
    def search(self, namespace, count, query):
        results = []

        features = self.encoder.encode(query)

        with self.lock:
            if namespace not in self.namespaces:
                logging.info('Namespace "{}" not indexed'.format(namespace))
                return results

            # Match query terms with embedings 
            for k,v in self.namespaces[namespace].items():
                # score is distance between 
                score = _cos_similarity(v, features)
                results.append((k, score))

        logging.info('{} results found'.format(len(results)))

        # sort results, in asc of score
        results.sort(key=lambda result: result[1], reverse=True)

        # return top #count# results
        return results[:count]

index = OnlineIndex()


# Default search engine
class OnlineSearchEngine():
    # request to index single entity
    # reference - unique way to refer
    # documents - all embeddings of entity
    #
    # returns: true, if new entry was added, false if existing entry was updated
    def index(self, namespace, uuid, body):

        # TODO: move this to queue client
        self.queue_host = '127.0.0.1'
        self.queue_port = 45000

        # TODO: move this to queue client
        with grpc.insecure_channel('{}:{}'.format(self.queue_host, self.queue_port)) as channel:
            stub = hawking_pb2_grpc.IndexStub(channel)
            logging.info('Add "{}/{}/{}"'.format(namespace, uuid, body))

            document = hawking_pb2.Document(uuid=uuid, body=body)
            request = hawking_pb2.IndexRequest(namespace=namespace, documents=[document])
            response = stub.Post(request)
            return response.inserted
            

    # search through entities (match query with embeddings of entity)
    # count - max number of results
    # query - search query
    #
    # returns: array of tuples (reference -> score)
    def search(self, namespace, count, query):
        return index.search(namespace, count, query)
        

engine = OnlineSearchEngine()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import sys
from setuptools import setup

platform = sys.platform
print(platform)

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = [
        'Click>=6.0',
        'bert-serving-server>=1.9.6',
        'bert-serving-client>=1.9.6',
        'flask',
        'flask_restplus',
        'numpy>=1.16.4',
        'tensorflow>=1.13',
        'tensorflow_hub>=0.5.0',
        'google-cloud-storage',
        'sentencepiece',
        'tf-sentencepiece',
        'pytorch-transformers',
        'python-telegram-bot>=12.0.0b1',
        'keras>=2.2.5',
        'Pillow',
        'hawking_proto']

setup_requirements = ['pytest-runner']

test_requirements = ['pytest']

setup(
    author="Peach Inc",
    author_email='hi@peach.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Operating System :: POSIX :: Linux",
        'Programming Language :: Python :: 3.6',
        "License :: OSI Approved :: Apache Software License"
    ],
    description="Hawking CLI",
    entry_points={
        'console_scripts': [
            'hawking=hawking.cli:main',
        ],
    },
    install_requires=requirements,
    keywords='hawking',
    name='hawking',
    packages=[
        'hawking',
        'hawking.encoder',
        'hawking.encoder.bert',
        'hawking.encoder.universal',
        'hawking.encoder.vgg',
        'hawking.encoder.resnet50',
        'hawking.encoder.inceptionresnetv2',
        'hawking.common',
        'hawking.serve',
        'hawking.frontend',
        'hawking.frontend.store',
        'hawking.frontend.api',
        'hawking.frontend.engine',
        'hawking.bot.telegram',
        'hawking.bot.secretary',
        'hawking.frontend'
    ],
    include_package_data=True,
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://www.peach.co',
    version='0.1.56',
    zip_safe=False,
    license='Apache License, Version 2.0'
)



import logging
import os
import time

from hawking.encoder.bert import OnlineBertTextEncoder, BertTextEncoder, encode_phrase
from hawking.encoder.inceptionresnetv2 import InceptionResNetV2ImageEncoder
from hawking.encoder.resnet50 import ResNet50ImageEncoder
from hawking.encoder.vgg import VGG16ImageEncoder, VGG19ImageEncoder
from hawking.encoder.universal import UniversalTextEncoder


def create_encoder(config={
                       "entity_type": "text",
                       "encoder_type": "universal-sentence-encoder",
                       "top_layer": "fc1000",
                       "print_summary": False,
                       "weights": "https://tfhub.dev/google/universal-sentence-encoder-multilingual-qa/1",
                       "encoder_host": "127.0.0.1",
                       "encoder_port": 5555,
                   }):
    """ Creating various encoder based on config.

    ResNet50 top-layer = 'fc1000'
    InceptionResNetV2 top-layer = 'predictions'
    """
    # TFHUB Caching 101: https://www.tensorflow.org/hub/basics.
    logging.info("Setting TFHUB_CACHE_DIR=/tmp.")
    os.environ["TFHUB_CACHE_DIR"] = "/tmp"
    st = time.time()
    logging.info(config)
    if config['entity_type'] == 'text':
        encoder_weights = config['weights']
        if config['encoder_type'] == 'online-bert':
            logging.info('Connecting to text encoder service `{}:{}`...'.format(config['encoder_host'],
                                                                                config['encoder_port']))
            encoder = OnlineBertTextEncoder(config['encoder_host'],
                                            config['encoder_port'])
        elif config['encoder_type'] == 'bert':
            encoder = BertTextEncoder(encoder_weights)
        elif config['encoder_type'] == 'universal-sentence-encoder':
            encoder = UniversalTextEncoder(encoder_weights)
        else:
            raise ValueError('encoder_weights `{}` is NOT supported.'.format(encoder_weights))
        logging.info('Successfully initialized text encoder with pre-trained weight `{}`.'.format(encoder_weights))
    elif config['entity_type'] == 'image':
        if config['encoder_type'] == 'resnet50':
            encoder = ResNet50ImageEncoder(top_layer=config['top_layer'], print_summary=config['print_summary'])
        elif config['encoder_type'] == 'inceptionresnetv2':
            encoder = InceptionResNetV2ImageEncoder(top_layer=config['top_layer'],
                                                    print_summary=config['print_summary'])
        elif config['encoder_type'] == 'vgg16':
            encoder = VGG16ImageEncoder(top_layer=config['top_layer'], print_summary=config['print_summary'])
        elif config['encoder_type'] == 'vgg19':
            encoder = VGG19ImageEncoder(top_layer=config['top_layer'], print_summary=config['print_summary'])
        else:
            raise ValueError('encoder_type:`{}` is not supported!'.format(config['encoder_type']))
    logging.info('Created encoder: %s. Elapsed time: %.2f' % (config['encoder_type'], time.time() - st))
    return encoder



import logging
import logging.config
import os
import time
import grpc

from hawking_proto import hawking_pb2, hawking_pb2_grpc

# Default search engine
class HawkingSearchEngine():
    def __init__(self):
        self.namespaces = {}
        self.indexer_host = '127.0.0.1'
        self.indexer_port = 50000
        self.channel = grpc.insecure_channel('{}:{}'.format(self.indexer_host, self.indexer_port))

    # request to index single entity
    # returns: true, if new entry was added, false if existing entry was updated
    def index(self, namespace, uuid, body):
        raise ValueError("Not supported!")

    # search through entities (match query with embeddings of entity)
    # count - max number of results
    # query - search query
    #
    # returns: array of tuples (reference -> score)
    def search(self, namespace, count, query, entity_type='text'):
        results = []

        with grpc.insecure_channel('{}:{}'.format(self.indexer_host, self.indexer_port)) as channel:
            stub = hawking_pb2_grpc.SearchStub(channel)

            st = time.time()
            type = (hawking_pb2.SearchRequest.EntityType.TEXT
                    if entity_type == 'text'
                    else hawking_pb2.SearchRequest.EntityType.IMAGE)
            logging.info('Type: {}. Query: "{}"'.format(type, query))
            request = hawking_pb2.SearchRequest(max_results=count, query=query, type=type)
            response = stub.Get(request)
            logging.info('Search results:\n')
            for result in response.results:
                snippet = '%s' % (result.document.body)
                results.append((snippet, result.score))
            logging.info(results)
            et = (time.time() - st) * 1000.0
            logging.info('elapsed time: %d ms' % et)

        return results


logging_conf_path = os.path.normpath(os.path.join(os.path.dirname(__file__), 'logging.conf'))
logging.config.fileConfig(logging_conf_path)
engine = HawkingSearchEngine()



""" Faiss Index Server.

"""
import os
import logging
import time

import click
import grpc

from concurrent import futures
from hawking_proto import hawking_pb2, hawking_pb2_grpc
from hawking.serve import configure_logging


_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class CaptionServicer(hawking_pb2_grpc.CaptionServicer):
    """ Implements hawking_pb2_grpc.CaptionServicer. """

    def __init__(self, index_path):
        self.LoadIndex(index_path)

    def LoadIndex(self, index_path):
        """ Load forward index. """
        assert os.path.exists(index_path), "Index {} does not exist!".format(index_path)
        logging.info('Loading index {}'.format(index_path))
        st = time.time()
        self.forward_index = []
        with open(index_path, 'r') as f:
            for line in f:
                line = line.strip()
                tokens = line.split('\t')
                assert len(tokens) == 1
                title = tokens[0]
                body = None
                if len(tokens) > 1:
                    body = tokens[1]
                meta = None
                if len(tokens) > 2:
                    meta = tokens[2]
                self.forward_index.append((title, body, meta))
        self.total_docs = len(self.forward_index)
        et = time.time() - st
        logging.info('Index loaded. Elapsed time: %.2f s' % et)

    def Get(self, request, context):
        logging.info('Receive query: {}'.format(request))

        response = hawking_pb2.SnippetResponse()
        for doc_id in request.doc_ids:
            title, body, meta = ('', '', '')
            if doc_id < self.total_docs:
                snippet = self.forward_index[doc_id]
                title, body, meta = snippet
            response.snippets.append(hawking_pb2.Snippet(title=title, body=body, metadata=meta))
        logging.info('Response:{}'.format(response))
        return response


def serve(index_path, worker_threads, grpc_port):
    configure_logging()

    logging.info('TAG: CAPTION')
    logging.info('Starting gRPC index server ...')
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=worker_threads))
    caption_servicer = CaptionServicer(index_path)
    hawking_pb2_grpc.add_CaptionServicer_to_server(caption_servicer, server)
    server.add_insecure_port('[::]:{}'.format(grpc_port))
    server.start()
    logging.info('Ready to serve. Listening on 0.0.0.0:{}'.format(grpc_port))
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


@click.command()
@click.option('--index-path', help="Index path.", default=None)
@click.option('--worker-threads', '-w', help="Maximum numbers of worker threads.", default=10)
@click.option('--grpc-port', help="gRPC port to listen to request.", default=50001)
def run(index_path, worker_threads, grpc_port):
    serve(index_path, int(worker_threads), int(grpc_port))


if __name__ == "__main__":
    run()



""" InceptionResNetV2 image encoder.

Returns the convolutional features of inception_resnet_v2.
"""
import logging
import numpy as np
import tensorflow as tf

from keras.preprocessing import image
from keras.applications.inception_resnet_v2 import InceptionResNetV2
from keras.applications.inception_resnet_v2 import preprocess_input
from keras.models import Model


class InceptionResNetV2ImageEncoder:
    def __init__(self, print_summary=True, top_layer='predictions', include_top=True):
        base_model = InceptionResNetV2(weights='imagenet', include_top=include_top)
        base_model.summary()
        self.model = Model(inputs=base_model.input, outputs=base_model.get_layer(top_layer).output)
        self.graph = tf.get_default_graph()
        if print_summary:
            self.model.summary()

    def batch_encode(self, img_paths):
        features = []
        for img_path in img_paths:
            features.append(self.encode(img_path))
        return features

    def encode(self, img_path, flattened=True):
        # Running predict on webservice, requires fetching tf.graph
        # See: https://github.com/keras-team/keras/issues/2397
        with self.graph.as_default():
            logging.info('Encode: {}'.format(img_path))
            img = image.load_img(img_path, target_size=(299, 299))
            img_data = image.img_to_array(img)
            img_data = np.expand_dims(img_data, axis=0)
            img_data = preprocess_input(img_data)
            feature = self.model.predict(img_data)
            feature_np = np.array(feature)
            return feature_np.flatten() if flattened else feature_np
